import copy
import abjad
from ._LoopParent import _LoopParent
from ..utilities.simplified_time_signature_ratio import (
    simplified_time_signature_ratio
)


class LoopByNotes(_LoopParent):
    r"""This class can be used to output slices of an ``abjad.Container`` using
    the metaphor of a looping window of a variable size. This size is given by
    the argument ``window_size``, which is an ``int`` representing how many
    notes are to be included in each slice. The duration of the slice will be
    the sum of the duration of these notes.

    For instance, if the initial container had the logical ties
    ``[A, B, C, D, E, F]`` (where each letter represents one logical tie) and
    the looping window was size ``3``, the output would be:

    ``A B C B C D C D E D E F E F F``

    This can be better visualised as:

    .. code-block:: none

        A B C
          B C D
            C D E
              D E F
                E F
                  F

    ..  container:: example

        Usage is similar to other factory classes. It takes a container (or
        child class equivalent) and the number of elements of the window as
        arguments. Each call of the object, in this case ``looper()``, will
        move the window forwards and output the result.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=3,
        ...                             )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByNotes-1.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }

        .. figure:: ../_images/image-LoopByNotes-2.png

        The property ``current_window`` can be used to access the current
        window without moving the head forwards.

        >>> notes = looper.current_window
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }

        .. figure:: ../_images/image-LoopByNotes-3.png

    ..  container:: example

        The very first call will output the input container without processing
        it. To disable this behaviour and have the looping window move on the
        very first call, initialise the class with the keyword argument
        ``move_window_on_first_call`` set to ``True``.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByNotes(
        ...     input_music,
        ...     window_size=3,
        ...     move_window_on_first_call=True,
        ... )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }

        .. figure:: ../_images/image-LoopByNotes-4.png

    ..  container:: example

        The instances of ``LoopByNotes`` can also be used as an iterator, which
        can then be used in a for loop to exhaust all windows.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=2,
        ...                             )
        >>> staff = abjad.Staff()
        >>> for window in looper:
        ...     staff.append(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'2
            d'2
            e'4
            \time 1/4
            e'4
        }

        .. figure:: ../_images/image-LoopByNotes-5.png

        Notice how the second staff in the example above does not have a time
        signature. This is because consecutive identical time signatures are
        omitted by default. To change this behaviour, instantialise this class
        with the keyword argument ``force_identical_time_signatures`` set to
        ``True``, or change the ``force_identical_time_signatures`` property
        to alter its value after the initialisation.

    ..  container:: example

        This class can take many optional keyword arguments during its
        creation. ``step_size`` dictates the size of each individual step in
        number of elements (default value is 1). ``max_steps`` sets the maximum
        number of steps that the window can advance when the object is called,
        ranging between 1 and the input value (default is also 1).
        ``repetition_chance`` sets the chance of a window result repeating
        itself (that is, the window not moving forwards when called). It should
        range from 0.0 to 1.0 (default 0.0, i.e. no repetition).
        ``forward_bias`` sets the chance of the window moving forward instead
        of backwards. It should range from 0.0 to 1.0 (default 1.0, which means
        the window can only move forwards. A value of 0.5 gives 50% chance of
        moving forwards while a value of 0.0 will move the window only
        backwards). Finally, ``head_position`` can be used to offset the
        starting position of the looping window. It must be an integer and its
        default value is 0.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=3,
        ...                             step_size=1,
        ...                             max_steps=2,
        ...                             repetition_chance=0.25,
        ...                             forward_bias=0.2,
        ...                             head_position=0,
        ...                             omit_all_time_signatures=False,
        ...                             force_identical_time_signatures=False,
        ...                             )
        >>> looper.window_size
        3
        >>> looper.step_size
        1
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        0
        >>> looper.omit_all_time_signatures
        False
        >>> looper.force_identical_time_signatures
        False

        Use the properties below to change these values after initialisation.

        >>> looper.window_size = 2
        >>> looper.step_size = 2
        >>> looper.max_steps = 3
        >>> looper.repetition_chance = 0.1
        >>> looper.forward_bias = 0.8
        >>> looper.head_position = 2
        >>> looper.omit_all_time_signatures = True
        >>> looper.force_identical_time_signatures = True
        >>> looper.window_size
        2
        >>> looper.step_size
        2
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        2
        >>> looper.omit_all_time_signatures
        True
        >>> looper.force_identical_time_signatures
        True

    ..  container:: example

        To disable time signatures altogether, initialise ``LoopByNotes`` with
        the keyword argument ``omit_all_time_signatures`` set to ``True``
        (default is ``False``), or use the ``omit_time_signature`` property
        after initialisation.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=3,
        ...                             omit_all_time_signatures=True,
        ...                             )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByNotes-6.png

    ..  container:: example

        The function ``len()`` can be used to get the total number of elements
        in the contents.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=3,
        ...                             )
        >>> len(looper)
        5

    ..  container:: example

        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method ``output_all()``.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=2,
        ...                             )
        >>> window = looper.output_all()
        >>> staff = abjad.Staff(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
            \time 2/4
            e'4
            f'4
            \time 1/4
            f'4
        }

        .. figure:: ../_images/image-LoopByNotes-7.png

    ..  container:: example

        When using ``output_all()``, set the keyword argument
        ``tie_identical_pitches`` to ``True`` in order to tie identical notes
        or chords at the end and beginning of consecutive windows.

        >>> input_music = abjad.Container(r"c'4 d'2 r8 d'4 <e' g'>8 r4 f'2. "
        ...                               "<e' g'>16")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=4,
        ...                             )
        >>> music = looper.output_all(tie_identical_pitches=True)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 9/8
            c'4
            d'2
            r8
            d'4
            ~
            \time 4/4
            d'2
            r8
            d'4
            <e' g'>8
            \time 3/4
            r8
            d'4
            <e' g'>8
            r4
            \time 11/8
            d'4
            <e' g'>8
            r4
            f'2.
            \time 19/16
            <e' g'>8
            r4
            f'2.
            <e' g'>16
            \time 17/16
            r4
            f'2.
            <e' g'>16
            \time 13/16
            f'2.
            <e' g'>16
            ~
            \time 1/16
            <e' g'>16
        }

        .. figure:: ../_images/image-LoopByNotes-8.png

    ..  container:: example

        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        ``output_n()`` and pass the number of iterations as argument. Similarly
        to ``output_all()``, the keyword argument ``tie_identical_pitches`` is
        available for tying pitches.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=2,
        ...                             )
        >>> window = looper.output_n(2)
        >>> staff = abjad.Staff(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
        }

        .. figure:: ../_images/image-LoopByNotes-9.png

    .. container:: example

        To change the size of the looping window after instantiation, use the
        property ``window_size``. In the example below, the initial window is
        of size 3, and so the first call of the looper object outputs the
        first, second, and third leaves. The window size is then set to 4, and
        the looper is called again, moving to the leaf in the next position,
        thus outputting the second, third, fourth, and fifth leaves.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=3,
        ...                             )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }

        .. figure:: ../_images/image-LoopByNotes-10.png

        >>> looper.window_size = 4
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 19/8
            d'2
            e'4
            f'2
            ~
            f'8
            g'1
        }

        .. figure:: ../_images/image-LoopByNotes-11.png

    .. container:: example

        Use the ``contents`` property to read as well as overwrite the contents
        of the looper. Notice that the ``head_position`` will remain on its
        previous value and must be reset to ``0`` if that's required.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
        >>> looper = auxjad.LoopByNotes(input_music,
        >>>                             window_size=3,
        >>>                             )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
        }

        .. figure:: ../_images/image-LoopByNotes-12.png

        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            d'4
            e'4
            f'4
        }

        .. figure:: ../_images/image-LoopByNotes-13.png

        >>> looper.contents = abjad.Container(r"c'''4 r4 d'''4 r4 "
        ...                                   "e'''4 r4 f'''4 r4")
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            d'''4
            r4
            e'''4
        }

        .. figure:: ../_images/image-LoopByNotes-14.png

        >>> looper.head_position = 0
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'''4
            r4
            d'''4
        }

        .. figure:: ../_images/image-LoopByNotes-15.png

    ..  container:: example

        This class can handle tuplets, but the output is not ideal and so this
        functionality should be considered experimental. Time signatures will
        be correct when dealing with partial tuplets (thus having non-standard
        values in their denominators), but each individual note of a tuplet
        will have the ratio printed above it.

        >>> input_music = abjad.Container(r"c'4 d'8 \times 2/3 {a4 g2}")
        >>> looper = auxjad.LoopByNotes(input_music,
        ...                             window_size=2,
        ...                             )
        >>> window = looper.output_all()
        >>> staff = abjad.Staff(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            #(ly:expect-warning "strange time signature found")
            \time 7/24
            d'8
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                \time 2/4
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                g2
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                #(ly:expect-warning "strange time signature found")
                \time 2/6
                g2
            }
        }

        .. figure:: ../_images/image-LoopByNotes-16.png
    """

    ### INITIALISER ###

    def __init__(self,
                 contents: abjad.Container,
                 *,
                 window_size: int,
                 step_size: int = 1,
                 max_steps: int = 1,
                 repetition_chance: float = 0.0,
                 forward_bias: float = 1.0,
                 head_position: int = 0,
                 omit_all_time_signatures: bool = False,
                 force_identical_time_signatures: bool = False,
                 move_window_on_first_call: bool = False,
                 ):
        self.contents = contents
        self._omit_all_time_signatures = omit_all_time_signatures
        self._force_identical_time_signatures = force_identical_time_signatures
        self._last_time_signature = None
        super().__init__(head_position,
                         window_size,
                         step_size,
                         max_steps,
                         repetition_chance,
                         forward_bias,
                         move_window_on_first_call,
                         )

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r'Outputs the representation of ``contents``.'
        return str(self._contents)

    def __len__(self) -> int:
        r'Outputs the number of logical ties of ``contents``.'
        return len(self._contents)

    ### PRIVATE METHODS ###

    def _slice_contents(self):
        r"""This method takes a slice with ``window_size`` number of logical
        ties out of the contents starting at the current ``head_position``.
        """
        start = self.head_position
        end = self.head_position + self.window_size
        logical_ties = self._contents[start:end]
        dummy_container = abjad.Container()
        time_signature_duration = 0
        for logical_tie in logical_ties:
            effective_duration = abjad.inspect(logical_tie).duration()
            logical_tie_ = copy.deepcopy(logical_tie)
            dummy_container.append(logical_tie_)
            multiplier = effective_duration / logical_tie_.written_duration
            logical_tie_ = abjad.mutate(logical_tie_).scale(multiplier)
            time_signature_duration += effective_duration
        if len(logical_ties) > 0 and not self._omit_all_time_signatures:
            time_signature = abjad.TimeSignature(time_signature_duration)
            time_signature = simplified_time_signature_ratio(time_signature)
            if (time_signature != self._last_time_signature
                    or self._force_identical_time_signatures):
                abjad.attach(time_signature,
                             abjad.select(dummy_container).leaves()[0],
                             )
            self._last_time_signature = time_signature
        self._current_window = dummy_container[:]
        dummy_container[:] = []

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self):
        r'The ``list`` which serves as the basis for the slices of the looper.'
        return self._contents

    @contents.setter
    def contents(self,
                 contents: abjad.Container,
                 ):
        if not isinstance(contents, abjad.Container):
            raise TypeError("'contents' must be 'abjad.Container' or "
                            "child class")
        contents_ = copy.deepcopy(contents)
        self._remove_all_time_signatures(contents_)
        self._contents = abjad.select(contents_).logical_ties()

    @property
    def omit_all_time_signatures(self) -> list:
        r'When ``True``, the output will contain no time signatures.'
        return self._omit_all_time_signatures

    @omit_all_time_signatures.setter
    def omit_all_time_signatures(self,
                                 omit_all_time_signatures: bool,
                                 ):
        if not isinstance(omit_all_time_signatures, bool):
            raise TypeError("'omit_all_time_signatures' must be 'bool'")
        self._omit_all_time_signatures = omit_all_time_signatures

    @property
    def force_identical_time_signatures(self) -> list:
        r"""When ``True``, identical time signatures will not be removed from
        the output.
        """
        return self._force_identical_time_signatures

    @force_identical_time_signatures.setter
    def force_identical_time_signatures(self,
                                        force_identical_time_signatures: bool,
                                        ):
        if not isinstance(force_identical_time_signatures, bool):
            raise TypeError("'force_identical_time_signatures' must be 'bool'")
        self._force_identical_time_signatures = force_identical_time_signatures



import random
import abjad
import pytest
import auxjad


def test_Shuffler_01():
    random.seed(87234)
    container = abjad.Container(r"c'4 d'4 e'4 f'4")
    assert repr(container) == 'Container("c\'4 d\'4 e\'4 f\'4")'
    shuffler = auxjad.Shuffler(container)
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            d'4
            c'4
            f'4
            e'4
        }
        """)
    music = shuffler.current_window
    with pytest.raises(AttributeError):
        shuffler.current_window = abjad.Container(r"c''2 e''2")
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            d'4
            c'4
            f'4
            e'4
        }
        """)


def test_Shuffler_02():
    container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 \time 2/4 f'4 g'4")
    shuffler = auxjad.Shuffler(container,
                               output_single_measure=False,
                               disable_rewrite_meter=False,
                               force_time_signatures=False,
                               )
    assert not shuffler.output_single_measure
    assert not shuffler.disable_rewrite_meter
    assert not shuffler.force_time_signatures
    shuffler.output_single_measure = True
    shuffler.disable_rewrite_meter = True
    shuffler.force_time_signatures = True
    assert shuffler.output_single_measure
    assert shuffler.disable_rewrite_meter
    assert shuffler.force_time_signatures


def test_Shuffler_03():
    random.seed(99961)
    container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 \time 2/4 f'4 g'4")
    shuffler = auxjad.Shuffler(container,
                               output_single_measure=True,
                               )
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 5/4
            f'4
            d'4
            e'4
            g'4
            c'4
        }
        """)


def test_Shuffler_04():
    random.seed(17453)
    container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 \time 2/4 f'2")
    shuffler = auxjad.Shuffler(container,
                               output_single_measure=True,
                               disable_rewrite_meter=True,
                               )
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 5/4
            d'4..
            f'2
            c'16
            e'4
        }
        """)


def test_Shuffler_05():
    random.seed(18892)
    container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 | r4 f'2")
    shuffler = auxjad.Shuffler(container)
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            d'4..
            e'16
            ~
            e'8.
            f'16
            ~
            f'4..
            r16
            r8.
            c'16
        }
        """)
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            c'16
            e'8.
            ~
            e'16
            f'4..
            ~
            f'16
            r8.
            r16
            d'4..
        }
        """)
    shuffler.force_time_signatures = True
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            d'4..
            r16
            r8.
            c'16
            f'2
            e'4
        }
        """)


def test_Shuffler_06():
    random.seed(98231)
    container = abjad.Container(r"\time 2/4 c'16 d'4.. | r4 e'8. f'16")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.output_n(3)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 2/4
            d'4..
            f'16
            c'16
            e'8.
            r4
            d'4..
            e'16
            ~
            e'8
            f'16
            r16
            r8.
            c'16
            r4
            d'4
            ~
            d'8.
            f'16
            c'16
            e'8.
        }
        """)


def test_Shuffler_07():
    random.seed(77347)
    container = abjad.Container(r"\time 3/4 c'16 d'4.. | r4 e'8. f'16")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.shuffle_pitches()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 3/4
        e'16
        c'4..
        r4
        d'8.
        f'16
    }
    """)


def test_Shuffler_08():
    random.seed(17231)
    container = abjad.Container(r"\times 2/3 {\time 5/4 c'4 d'2} r4 e'4. f'8")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.output_n_shuffled_pitches(3)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \times 2/3 {
            \time 5/4
            f'4
            e'2
        }
        r4
        d'4.
        c'8
        \times 2/3 {
            d'4
            c'2
        }
        r4
        f'4.
        e'8
        \times 2/3 {
            d'4
            f'2
        }
        r4
        c'4.
        e'8
    }
    """)


def test_Shuffler_09():
    container = abjad.Container(r"\time 3/4 c'16 d'4.. | r4 e'8. f'16")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.rotate_pitches()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 3/4
        d'16
        e'4..
        r4
        f'8.
        c'16
    }
    """)


def test_Shuffler_10():
    container = abjad.Container(r"\time 3/4 c'16 d'4.. | r4 e'8. f'16")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.rotate_pitches(anticlockwise=True, n_rotations=2)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \time 3/4
        e'16
        f'4..
        r4
        c'8.
        d'16
    }
    """)


def test_Shuffler_11():
    container = abjad.Container(r"\times 2/3 {\time 5/4 c'4 d'2} r4 e'4. f'8")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.output_n_rotated_pitches(3)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
    r"""
    \new Staff
    {
        \times 2/3 {
            \time 5/4
            d'4
            e'2
        }
        r4
        f'4.
        c'8
        \times 2/3 {
            e'4
            f'2
        }
        r4
        c'4.
        d'8
        \times 2/3 {
            f'4
            c'2
        }
        r4
        d'4.
        e'8
    }
    """)


def test_Shuffler_12():
    random.seed(87234)
    container = abjad.Container(r"c'4 d'4 e'4 f'4")
    shuffler = auxjad.Shuffler(container)
    music = shuffler.shuffle()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 4/4
            d'4
            c'4
            f'4
            e'4
        }
        """)


def test_Shuffler_13():
    random.seed(18892)
    container = abjad.Container(r"\time 3/4 c'16 d'4.. e'4 | r4 f'2")
    shuffler = auxjad.Shuffler(container,
                               omit_time_signatures=True,
                               )
    music = shuffler()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            d'4..
            e'16
            ~
            e'8.
            f'16
            ~
            f'4..
            r16
            r8.
            c'16
        }
        """)
    assert shuffler.omit_time_signatures
    shuffler.omit_time_signatures = False
    assert not shuffler.omit_time_signatures



def test_Shuffler_12():
    random.seed(47621)
    container = abjad.Container(r"c'4 d'4 e'4 f'4")
    shuffler = auxjad.Shuffler(container)
    assert format(shuffler.contents) == abjad.String.normalize(
        r"""
        {
            c'4
            d'4
            e'4
            f'4
        }
        """)
    shuffler()
    assert format(shuffler.contents) == abjad.String.normalize(
        r"""
        {
            c'4
            d'4
            e'4
            f'4
        }
        """)
    shuffler.contents = abjad.Container(r"cs2 ds2")
    assert format(shuffler.contents) == abjad.String.normalize(
        r"""
        {
            cs2
            ds2
        }
        """)



import abjad


def leaves_are_tieable(leaf1: abjad.Leaf,
                       leaf2: abjad.Leaf,
                       ) -> bool:
    r"""Returns a ``bool`` representing whether or not two input leaves (of
    type ``abjad.Leaf`` or child class) have identical pitch(es) and thus can
    be tied.

    ..  container:: example

        When the pitches in both leaves are identical, this function returns
        ``True``:

        >>> Leaf1 = abjad.Note(r"c'4")
        >>> Leaf2 = abjad.Note(r"c'4")
        >>> auxjad.leaves_are_tieable(Leaf1, Leaf2)
        True

    ..  container:: example

        Durations do not affect the comparison.

        >>> Leaf1 = abjad.Note(r"c'2.")
        >>> Leaf2 = abjad.Note(r"c'16")
        >>> Leaf3 = abjad.Note(r"f'''16")
        >>> auxjad.leaves_are_tieable(Leaf1, Leaf2)
        True
        >>> auxjad.leaves_are_tieable(Leaf1, Leaf3)
        False
        >>> auxjad.leaves_are_tieable(Leaf2, Leaf3)
        False

    ..  container:: example

        Handles chords as well as pitches.

        >>> chord1 = abjad.Chord(r"<c' e' g'>4")
        >>> chord2 = abjad.Chord(r"<c' e' g'>16")
        >>> chord3 = abjad.Chord(r"<f''' fs'''>16")
        >>> auxjad.leaves_are_tieable(chord1, chord2)
        True
        >>> auxjad.leaves_are_tieable(chord1, chord3)
        False
        >>> auxjad.leaves_are_tieable(chord2, chord3)
        False

    ..  container:: example

        Leaves can also be part of containers.

        >>> container = abjad.Container(r"r4 <c' e'>4 <c' e'>2")
        >>> auxjad.leaves_are_tieable(container[1], container[2])
        True

    ..  container:: example

        If rests are input, the return value is ``False``.

        >>> container = abjad.Container(r"r4 g'4 r2")
        >>> auxjad.leaves_are_tieable(container[0], container[2])
        False
    """
    if not isinstance(leaf1, abjad.Leaf):
        raise TypeError("first positional argument must be 'abjad.Leaf' or "
                        "child class")
    if not isinstance(leaf2, abjad.Leaf):
        raise TypeError("second positional argument must be 'abjad.Leaf' or "
                        "child class")
    if type(leaf1) is not type(leaf2):
        return False
    if isinstance(leaf1, abjad.Rest):
        return False
    if isinstance(leaf1, abjad.MultimeasureRest):
        return False
    if (isinstance(leaf1, abjad.Note) and 
            leaf1.written_pitch != leaf2.written_pitch):
        return False
    if (isinstance(leaf1, abjad.Chord) and 
            leaf1.written_pitches != leaf2.written_pitches):
        return False
    if (type(abjad.inspect(leaf1).before_grace_container()) is not 
            type(abjad.inspect(leaf2).before_grace_container())):
        return False
    return True



import abjad


def respell_chord(chord: abjad.Chord,
                  *,
                  include_multiples: bool = False,
                  respell_by_pitch_class: bool = False,
                  ):
    r"""Mutates an input chord (of type ``abjad.Chord`` or child class) in
    place and has no return value. This function changes the accidentals of
    individual pitches of a chord in order to avoid augmented unisons.

    ..  container:: example

        To use this function, apply it to a chord that contains augmented
        unisons.

        >>> chord = abjad.Chord("<c' cs'>4")
        >>> auxjad.respell_chord(chord)
        >>> abjad.f(chord)
        <c' df'>4

        .. figure:: ../_images/image-respell_chord-1.png

    ..  container:: example

        The example below shows the default spelling of 2-note chords by
        ``Abjad`` in the upper staff, and the respelt 2-note chords in the
        bottom staff.

        >>> staff1 = abjad.Staff()
        >>> staff2 = abjad.Staff()
        >>> for pitch in range(12):
        ...     staff1.append(abjad.Chord([pitch, pitch + 1], (1, 16)))
        ...     chord = abjad.Chord([pitch, pitch + 1], (1, 16))
        ...     auxjad.respell_chord(chord)
        ...     staff2.append(chord)
        >>> literal = abjad.LilyPondLiteral(r'\accidentalStyle dodecaphonic')
        >>> abjad.attach(literal, staff1)
        >>> abjad.attach(literal, staff2)
        >>> score = abjad.Score([staff1, staff2])
        >>> abjad.f(score)
        \new Score
        <<
            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' cs'>16
                <cs' d'>16
                <d' ef'>16
                <ef' e'>16
                <e' f'>16
                <f' fs'>16
                <fs' g'>16
                <g' af'>16
                <af' a'>16
                <a' bf'>16
                <bf' b'>16
                <b' c''>16
            }
            \new Staff
            {
                \accidentalStyle dodecaphonic
                <c' df'>16
                <cs' d'>16
                <d' ef'>16
                <ds' e'>16
                <e' f'>16
                <f' gf'>16
                <fs' g'>16
                <g' af'>16
                <gs' a'>16
                <a' bf'>16
                <as' b'>16
                <b' c''>16
            }
        >>

        .. figure:: ../_images/image-respell_chord-2.png

    ..  container:: example

        The function looks for all augmented unissons in chords of 3 or more
        pitches:

        >>> chord1 = abjad.Chord(r"<a c' cs' f'>1")
        >>> chord2 = abjad.Chord(r"<a c' cs' f'>1")
        >>> auxjad.respell_chord(chord2)
        >>> staff = abjad.Staff([chord1, chord2])
        >>> abjad.f(staff)
        \new Staff
        {
            <a c' cs' f'>1
            <a c' df' f'>1
        }

        .. figure:: ../_images/image-respell_chord-3.png

        It is not a problem if the pitches are input out of order.

        >>> chord1 = abjad.Chord(r"<e' cs' g' ef'>1")
        >>> chord2 = abjad.Chord(r"<e' cs' g' ef'>1")
        >>> auxjad.respell_chord(chord2)
        >>> staff = abjad.Staff([chord1, chord2])
        >>> abjad.f(staff)
        \new Staff
        {
            <cs' ef' e' g'>1
            <cs' ds' e' g'>1
        }

        .. figure:: ../_images/image-respell_chord-4.png

    ..  container:: example

        By default, this function only changes spelling for pitches that are
        1 semitone apart.

        >>> chord1 = abjad.Chord(r"<c' cs''>1")
        >>> chord2 = abjad.Chord(r"<c' cs''>1")
        >>> auxjad.respell_chord(chord2)
        >>> staff = abjad.Staff([chord1, chord2])
        >>> abjad.f(staff)
        \new Staff
        {
            <c' cs''>1
            <c' cs''>1
        }

        .. figure:: ../_images/image-respell_chord-5.png

        To consider pitches in different octaves (thus including augmented
        unisons, augmented octaves, augmented fifteenths, etc.), call this
        function with the keyword argument ``include_multiples`` set to
        ``True``.

        >>> chord1 = abjad.Chord(r"<c' cs''>1")
        >>> chord2 = abjad.Chord(r"<c' cs''>1")
        >>> auxjad.respell_chord(chord2, include_multiples=True)
        >>> staff = abjad.Staff([chord1, chord2])
        >>> abjad.f(staff)
        \new Staff
        {
            <c' cs''>1
            <c' df''>1
        }

        .. figure:: ../_images/image-respell_chord-6.png

    ..  container:: example

        By default, when this function changes the spelling of a pitch, it does
        not change the spelling of all other pitches with the same pitch-class.

        >>> chord1 = abjad.Chord(r"<c' cs' cs''>1")
        >>> chord2 = abjad.Chord(r"<c' cs' cs''>1")
        >>> auxjad.respell_chord(chord2)
        >>> staff = abjad.Staff([chord1, chord2])
        >>> abjad.f(staff)
        \new Staff
        {
            <c' cs' cs''>1
            <c' df' cs''>1
        }

        .. figure:: ../_images/image-respell_chord-7.png

        To alter all pitch-classes, call this function with the keyword
        argument ``respell_by_pitch_class`` set to ``True``.

        >>> chord1 = abjad.Chord(r"<c' cs' cs''>1")
        >>> chord2 = abjad.Chord(r"<c' cs' cs''>1")
        >>> auxjad.respell_chord(chord2, respell_by_pitch_class=True)
        >>> staff = abjad.Staff([chord1, chord2])
        >>> abjad.f(staff)
        \new Staff
        {
            <c' cs' cs''>1
            <c' df' df''>1
        }

        .. figure:: ../_images/image-respell_chord-8.png
    """
    if not isinstance(chord, abjad.Chord):
        raise TypeError("argument must be 'abjad.Chord'")
    if not isinstance(respell_by_pitch_class, bool):
        raise TypeError("'respell_by_pitch_class' must be 'bool'")
    if not isinstance(include_multiples, bool):
        raise TypeError("'include_multiples' must be 'bool'")

    pitch_accidentals = [pitch.accidental for pitch in chord.written_pitches]

    if not include_multiples:
        # dealing only with intervals of size equal to 1 semitone
        for i in range(len(chord.written_pitches) - 1):
            p1 = chord.written_pitches[i]
            p2 = chord.written_pitches[i + 1]
            interval12 = p1 - p2
            try:
                p3 = chord.written_pitches[i + 2]
                interval23 = p2 - p3
            except IndexError:
                p3 = None
                interval23 = None
            if (interval12 == abjad.NamedInterval('+A1')
                    and interval23 != abjad.NamedInterval('+A1')
                    and interval23 != abjad.NamedInterval('+m2')):
                if not respell_by_pitch_class:
                    # respelling only one single note
                    if p1.accidental == abjad.Accidental('f'):
                        pitch_accidentals[i] = abjad.Accidental('s')
                    elif p1.accidental == abjad.Accidental(''):
                        pitch_accidentals[i + 1] = abjad.Accidental('f')
                else:
                    # respelling all matching pitch-classes
                    if p1.accidental == abjad.Accidental('f'):
                        for k, p in enumerate(chord.written_pitches):
                            if p.pitch_class == p1.pitch_class:
                                pitch_accidentals[k] = abjad.Accidental('s')
                    elif p1.accidental == abjad.Accidental(''):
                        for k, p in enumerate(chord.written_pitches):
                            if p.pitch_class == p2.pitch_class:
                                pitch_accidentals[k] = abjad.Accidental('f')
    else:
        # dealing with augmented unisons as well as augmented 8as, 15ths, etc.
        for i in range(len(chord.written_pitches) - 1):
            for j in range(i + 1, len(chord.written_pitches)):
                # no p3 this time since p1 and p2 are not necessary consecutive
                # pitches in the chord
                p1 = chord.written_pitches[i]
                p2 = chord.written_pitches[j]
                interval = abjad.NamedIntervalClass(p1 - p2)
                if (interval == abjad.NamedIntervalClass('+A1')
                        or interval == abjad.NamedIntervalClass('-d1')):
                    # no need for respell_by_pitch_class since this will go
                    # through all notes in the chord anyway
                    if p1.accidental == abjad.Accidental('f'):
                        pitch_accidentals[i] = abjad.Accidental('s')
                    if p1.accidental == abjad.Accidental('s'):
                        pitch_accidentals[i] = abjad.Accidental('f')
                    elif p1.accidental == abjad.Accidental(''):
                        if p2.accidental == abjad.Accidental('s'):
                            pitch_accidentals[j] = abjad.Accidental('f')
                        elif p2.accidental == abjad.Accidental('f'):
                            pitch_accidentals[j] = abjad.Accidental('s')

    # rewritting chord with new spelling
    respelt_pitches = []
    for pitch, accidental in zip(chord.written_pitches, pitch_accidentals):
        if accidental == abjad.Accidental('f'):
            respelt_pitches.append(pitch._respell_with_flats())
        elif accidental == abjad.Accidental('s'):
            respelt_pitches.append(pitch._respell_with_sharps())
        else:
            respelt_pitches.append(pitch)
    chord.written_pitches = respelt_pitches



import abjad
import auxjad


def test_rests_to_multimeasure_rest_01():
    container = abjad.Container(r"r1")
    auxjad.rests_to_multimeasure_rest(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            R1
        }
        """)


def test_rests_to_multimeasure_rest_02():
    container = abjad.Container(r"\time 3/4 r4 r8.. r32 r4")
    auxjad.rests_to_multimeasure_rest(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            %%% \time 3/4 %%%
            R1 * 3/4
        }
        """)


def test_rests_to_multimeasure_rest_03():
    container = abjad.Container(r"\time 3/4 r2. | "
                                r"\time 6/8 r2. | "
                                r"\time 5/4 c'1 ~ c'4 | r1 r4"
                                )
    auxjad.rests_to_multimeasure_rest(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            %%% \time 3/4 %%%
            R1 * 3/4
            %%% \time 6/8 %%%
            R1 * 3/4
            %%% \time 5/4 %%%
            c'1
            ~
            c'4
            R1 * 5/4
        }
        """)



import random


class TenneySelector():
    r"""This in an implementation of the Dissonant Counterpoint Algorithm by
    James Tenney. This class can be used to randomly select elements from an
    input list, giving more weight to elements which have not been selected in
    recent iterations. In other words, Tenney's algorithm uses feedback in
    order to lower the weight of recently selected elements.

    This implementation is based on the paper: Polansky, L., A. Barnett, and
    M. Winter (2011). 'A Few More Words About James Tenney: Dissonant
    Counterpoint and Statistical Feedback'. In: Journal of Mathematics and
    Music 5(2). pp. 63--82.

    ..  container:: example

        The selector should be initialised with a list of objects. The contents
        of the list can be absolutely anything.

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> selector.contents
        ['A', 'B', 'C', 'D', 'E', 'F']

        Applying the ``len()`` function to the selector will give the length
        of the input list.

        >>> len(selector)
        6

        When no other keyword arguments are used, the default probabilities of
        each element in the list is 1.0. Probabilities are not normalised. Use
        the previous_index attribute to check the previously selected index
        (default is None).

        >>> selector.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.previous_index
        None

        Calling the selector will output one of its elements, selected
        according to the current probability values.

        >>> selector()
        C

        Alternatively, use the ``next()`` function or ``__next__()`` method to
        get the next result.

        >>> selector.__next__()
        A
        >>> next(selector)
        D

        After each call, the object updates all probability values, setting the
        previously selected element's probability at 0.0 and raising all other
        probabilities according to a growth function (more on this below).

        >>> result = ''
        >>> for _ in range(30):
        ...     result += selector()
        >>> result
        EDFACEABAFDCEDAFADCBFEDABEDFEC

        From the result above it is possible to see that there are no immediate
        repetitions of elements (since once selected, their probability is
        always set to 0.0 and will take at least one iteration to grow to a
        non-zero value). Checking the probabilities and previous_index
        attributes will give us their current values.

        >>> selector.probabilities
        [6.0, 5.0, 0.0, 3.0, 1.0, 2.0]
        >>> selector.previous_index
        2

    ..  container:: example

        This class can take two optional keywords argument during its
        instantiation, namely ``weights`` and ``curvature``. ``weights`` takes
        a list of floats with the individual weights of each element; by
        default,  all weights are set to 1.0. These weights affects the
        effective probability of each element. The other argument,
        ``curvature``, is the exponent of the growth function for all elements.
        The growth function takes as input the number of iterations since an
        element has been last selected, and raise this number by the curvature
        value. If ``curvature`` is set to 1.0 (which is its default value), the
        growth is linear with each iteration. If set to a value larger than 0.0
        and less than 1.0, the growth is negative (or concave), so that the
        chances of an element which is not being selected will grow at ever
        smaller rates as the number of iterations it has not been selected
        increase. If the ``curvature`` is set to 1.0, the growth is linear with
        the number of iterations. If the ``curvature`` is larger than 1.0, the
        curvature is positive (or convex) and the growth will accelerate as the
        number of iterations an element has not been selected grows. Setting
        the curvature to 0.0 will result in an static probability vector with
        all values set to 1.0, except for the previously selected one which
        will be set to 0.0; this will result in a uniformly random selection
        without repetition.

        With linear curvature (default value of 1.0):

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> selector.curvature
        1.0
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector()
        'B'
        >>> selector.curvature
        1.0
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [2.0, 0.0, 2.0, 2.0, 2.0, 2.0]

    ..  container:: example

        Using a convex curvature:

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'],
        ...                                  curvature=0.2,
        ...                                  )
        >>> selector.curvature
        0.2
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector()
        'C'
        >>> selector.curvature
        0.2
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [1.148698354997035, 1.148698354997035, 0.0, 1.148698354997035,
        1.148698354997035, 1.148698354997035]

        With a convex curvature, the growth of the probability of each
        non-selected term gets smaller as the number of times it is not
        selected increases. The smaller the curvature is, the less difference
        there will be between any non-previously selected elements. This
        results in sequences which have more chances of a same element being
        near each other. In the sequence below, note how there are many cases
        of a same element being separated only by a single other one, such as
        ``'ACA'`` in index 6.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += selector()
        >>> result
        DACBEDFACABDACECBEFAEDBAFBABFD

        Checking the probability values at this point outputs:

        >>> selector.probabilities
        [1.2457309396155174, 1.148698354997035, 1.6952182030724354, 0.0,
        1.5518455739153598, 1.0]

        As we can see, all non-zero values are relatively close to each other,
        which is why there is a high chance of an element being selected again
        just two iterations apart.

    ..  container:: example

        Using a concave curvature:

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'],
        ...                                  curvature=15.2,
        ...                                  )
        >>> selector.curvature
        0.2
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector()
        'C'
        >>> selector.curvature
        0.2
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [37640.547696542824, 37640.547696542824, 37640.547696542824, 0.0,
        37640.547696542824, 37640.547696542824]

        With a concave curvature, the growth of the probability of each
        non-selected term gets larger as the number of times it is not selected
        increases. The larger the curvature is, the larger difference there
        will be between any non-previously selected elements. This results in
        sequences which have less chances of a same element being near each
        other. In the sequence below, with a curvature of 15.2, note how the
        elements are as far apart from each other, resulting in a repeating
        string of ``'DFAECB'``.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += selector()
        >>> result
        DFAECBDFAECBDFAECBDFAECBDFAECB

        Checking the probability values at this point outputs:

        >>> selector.probabilities
        [17874877.39956566, 0.0, 1.0, 42106007735.02238,
        37640.547696542824, 1416810830.8957152]

        As we can see, the non-zero values vary wildly. The higher the
        curvature, the higher the difference between these values, making some
        of them much more likely to be selected.

    ..  container:: example

        Each element can also have a fixed weight to themselves. This will
        affect the probability calculation. The example below uses the default
        linear curvature.

        >>> selector = auxjad.TenneySelector(
        ...     ['A', 'B', 'C', 'D', 'E', 'F'],
        ...     weights=[1.0, 1.0, 5.0, 5.0, 10.0, 20.0],
        >>> )
        >>> selector.weights
        [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
        >>> selector.probabilities
        [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
        >>> result = ''
        >>> for _ in range(30):
        ...     result += selector()
        >>> result
        FBEFECFDEADFEDFEDBFECDAFCEDCFE
        >>> selector.weights
        [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
        >>> selector.probabilities
        [7.0, 12.0, 10.0, 15.0, 0.0, 20.0]

    ..  container:: example

        To reset the probability distribution of all elements to its initial
        value (an uniform distribution), use the method
        ``reset_probabilities()``.

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> for _ in range(30):
        ...     selector()
        >>> selector.probabilities
        [4.0, 3.0, 1.0, 0.0, 5.0, 2.0]
        >>> selector.reset_probabilities()
        >>> selector.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]

    ..  container:: example

        This class allows slicing to get and set values of contents of the
        selector. This will not affect the current probability vector, and the
        new element will have the same probability as the one it replaced.

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> for _ in range(30):
        ...     selector()
        >>> selector.probabilities
        [3.0, 2.0, 1.0, 7.0, 5.0, 0.0]
        >>> selector[2]
        'C'
        >>> selector[1:4]
        ['B', 'C', 'D']
        >>> selector[2] = 'foo'
        >>> selector.contents
        ['A', 'B', 'foo', 'D', 'E', 'F']
        >>> selector[:] = ['foo', 'bar', 'X', 'Y', 'Z', '...']
        >>> selector.contents
        ['foo', 'bar', 'X', 'Y', 'Z', '...']
        >>> selector.probabilities
        [3.0, 2.0, 1.0, 7.0, 5.0, 0.0]

        You can also check if the instance contains a specific element. In the
        case of the selector above, we have:

        >>> 'foo' in selector
        True
        >>> 'A' in selector
        False

    ..  container:: example

        A new list of an arbitrary length can be set at any point using the
        property ``contents``. Do notice that the probabilities will be reset
        at that point. This method can take the optional keyword argument
        weights similarly to when instantiating the class.

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> for _ in range(30):
        ...     selector()
        >>> selector.probabilities
        [2.0, 1.0, 4.0, 3.0, 0.0, 5.0]
        >>> selector.contents
        ['A', 'B', 'C', 'D', 'E', 'F']
        >>> selector.contents = [2, 4, 6, 8]
        >>> selector.contents
        [2, 4, 6, 8]
        >>> len(selector)
        4
        >>> selector.weights
        [1.0, 1.0, 1.0, 1.0]
        >>> selector.probabilities
        [1.0, 1.0, 1.0, 1.0]

    ..  container:: example

        To change the curvature value at any point, simply set the property
        ``curvature`` to a different value.

        >>> selector = auxjad.TenneySelector(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> selector.curvature
        1.0
        >>> selector.curvature = 0.25
        >>> selector.curvature
        0.25
    """

    ### INITIALISER ###

    def __init__(self,
                 contents: list,
                 *,
                 weights: list = None,
                 curvature: float = 1.0,
                 ):
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list'")
        if weights:
            if not isinstance(weights, list):
                raise TypeError("'weights' must be 'list'")
            if not len(contents) == len(weights):
                raise ValueError("'weights' must have the same length "
                                 "as 'contents'")
            if not all(isinstance(weight, (int, float))
                       for weight in weights):
                raise TypeError("'weights' elements must be "
                                "'int' or 'float'")
        if not isinstance(curvature, float):
            raise TypeError("'curvature' must be 'float'")
        if curvature < 0.0:
            raise ValueError("'curvature' must be larger than 0.0")

        self._contents = contents[:]
        if weights:
            self._weights = weights[:]
        else:
            self._weights = [1.0 for _ in range(self.__len__())]
        self._curvature = curvature
        self._counter = [1 for _ in range(self.__len__())]
        self._generate_probabilities()
        self._previous_index = None

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r'Outputs the representation of ``contents``.'
        return str(self._contents)

    def __call__(self):
        r'Call the selection process and outputs one element of ``contents``.'
        self._previous_index = random.choices(
            [n for n in range(self.__len__())],
            weights=self.probabilities,
            )[0]
        self._regenerate_counts()
        self._generate_probabilities()
        return self._contents[self._previous_index]

    def __next__(self):
        r'Call the selection process and outputs one element of ``contents``.'
        return self.__call__()

    def __len__(self) -> int:
        r'Outputs the length of ``contents``.'
        return len(self._contents)

    def __getitem__(self, key: int):
        r"""Implements reading elements of ``contents`` through indexing or
        slicing of instance.
        """
        return self._contents[key]

    def __setitem__(self, key, value):
        r"""Implements writing elements into ``contents`` through indexing or
        slicing of instance.
        """
        self._contents[key] = value

    ### PUBLIC METHODS ###

    def reset_probabilities(self):
        r"""Resets the probability distribution of all elements to an uniform
        distribution.
        """
        self._counter = [1 for _ in range(self.__len__())]
        self._generate_probabilities()

    ### PRIVATE METHODS ###

    def _regenerate_counts(self):
        r"""Increases the count of all elements except for the previously
        selected one, whose count is reset to zero.
        """
        for i in range(self.__len__()):
            if i == self._previous_index:
                self._counter[i] = 0
            else:
                self._counter[i] += 1

    def _generate_probabilities(self,
                                reset: bool = False,
                                ):
        r"""Generates the probabilities given the weights of the elements as
        well as their count numbers (which are fed into the growth function).
        """
        if reset:
            self._counter = [1 for _ in range(self.__len__())]
        self.probabilities = []
        for weight, count in zip(self._weights, self._counter):
            self.probabilities.append(weight * self._growth_function(count))

    def _growth_function(self, count):
        r'Applies the growth exponent given a number of counts.'
        return count ** self._curvature

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> list:
        r'The ``list`` from which the selector picks elements.'
        return self._contents

    @contents.setter
    def contents(self,
                 contents: list,
                 ):
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list")
        self._contents = contents[:]
        self.weights = [1.0 for _ in range(self.__len__())]

    @property
    def weights(self) -> list:
        r'The ``list`` with weights for each element of ``contents``.'
        return self._weights

    @weights.setter
    def weights(self,
                weights: list,
                ):
        if not isinstance(weights, list):
            raise TypeError("'weights' must be 'list'")
        if not self.__len__() == len(weights):
            raise ValueError("'weights' must have the same length "
                             "as the contents of the object")
        if not all(isinstance(weight, (int, float))
                   for weight in weights):
            raise TypeError("'weights' elements must be "
                            "'int' or 'float'")
        self._weights = weights[:]
        self._generate_probabilities(reset=True)

    @property
    def curvature(self) -> list:
        r'The exponent of the growth function.'
        return self._curvature

    @curvature.setter
    def curvature(self,
                  curvature: float,
                  ):
        if not isinstance(curvature, float):
            raise TypeError("'curvature' must be 'float'")
        if curvature < 0.0:
            raise ValueError("'curvature' must be larger than 0.0")
        self._curvature = curvature
        self._generate_probabilities()

    @property
    def previous_index(self):
        r"""Read-only property, returns the index of the previously output
        element.
        """
        return self._previous_index



import abjad


class LeafDynMaker(abjad.LeafMaker):
    r"""This class can be used to create leaves and logical ties from input
    lists of pitches, durations, dynamics, and articulations. It is an
    extension of ``abjad.LeafMaker`` which can take optional lists of dynamics
    and articulations.

    ..  container:: example

        Usage is similar to ``abjad.LeafMaker``:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
        >>> dynamics = ['pp', 'p', 'mp', 'mf', 'f', 'ff']
        >>> articulations = ['.', '>', '-', '_', '^', '+']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'32
            \pp
            -\staccato
            d'16
            \p
            -\accent
            e'16.
            \mp
            -\tenuto
            f'8
            \mf
            -\portato
            g'8
            \f
            -\marcato
            ~
            g'32
            a'8.
            \ff
            -\stopped
        }

        .. figure:: ../_images/image-LeafDynMaker-1.png

    ..  container:: example

        Tuple elements in ``pitches`` result in chords. None-valued elements
        in ``pitches`` result in rests:

        >>> pitches = [5, None, (0, 2, 7)]
        >>> durations = [(1, 4), (1, 8), (1, 16)]
        >>> dynamics = ['p', None, 'f']
        >>> articulations = ['staccato', None, 'tenuto']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            f'4
            \p
            -\staccato
            r8
            <c' d' g'>16
            \f
            -\tenuto
        }

        .. figure:: ../_images/image-LeafDynMaker-2.png

    ..  container:: example

        Can omit repeated dynamics with the keyword argument
        ``omit_repeated_dynamics``:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = [(1, 32), (2, 32), (3, 32), (4, 32), (5, 32), (6, 32)]
        >>> dynamics = ['pp', 'pp', 'mp', 'f', 'f', 'p']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches,
        ...                        durations,
        ...                        dynamics,
        ...                        omit_repeated_dynamics=True,
        ...                        )
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'32
            \pp
            d'16
            e'16.
            \mp
            f'8
            \f
            g'8
            ~
            g'32
            a'8.
            \p
        }

        .. figure:: ../_images/image-LeafDynMaker-3.png

    ..  container:: example

        The lengths ``dynamics`` and ``articulations`` can be shorter than the
        lengths of ``pitches`` and ``durations`` (whatever is the greatest):

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = (1, 4)
        >>> dynamics = ['p', 'f', 'ff']
        >>> articulations = ['.', '>']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            -\staccato
            d'4
            \f
            -\accent
            e'4
            \ff
            f'4
            g'4
            a'4
        }

        .. figure:: ../_images/image-LeafDynMaker-4.png

    ..  container:: example

        If the lengths of either ``dynamics`` and ``articulations`` are shorter
        than the lengths of ``pitches`` and ``durations`` (whatever is the
        greatest), use the optional keyword arguments ``cyclic_dynamics`` and
        ``cyclic_articulations`` to apply those parameters cyclically:

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = (1, 4)
        >>> dynamics = ['p', 'f', 'ff']
        >>> articulations = ['.', '>']
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches,
        ...                        durations,
        ...                        dynamics,
        ...                        articulations,
        ...                        cyclic_dynamics=True,
        ...                        cyclic_articulations=True,
        ...                        )
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            - \staccato
            d'4
            \f
            - \accent
            e'4
            \ff
            - \staccato
            f'4
            \p
            - \accent
            g'4
            \f
            - \staccato
            a'4
            \ff
            - \accent
        }

        .. figure:: ../_images/image-LeafDynMaker-5.png

    ..  container:: example

        If the length of ``articulations`` or ``dynamics`` is 1, they will be
        applied only to the first element.

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = (1, 4)
        >>> dynamics = 'p'
        >>> articulations = '.'
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            -\staccato
            d'4
            e'4
            f'4
            g'4
            a'4
        }

        .. figure:: ../_images/image-LeafDynMaker-6.png

    ..  container:: example

        To apply them to all elements, use the ``cyclic_dynamics`` and
        ``cyclic_articulations`` optioanl keywords.

        >>> pitches = [0, 2, 4, 5, 7, 9]
        >>> durations = (1, 4)
        >>> dynamics = 'p'
        >>> articulations = '.'
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches,
        ...                        durations,
        ...                        dynamics,
        ...                        articulations,
        ...                        cyclic_articulations=True,
        ...                        )
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            -\staccato
            d'4
            -\staccato
            e'4
            -\staccato
            f'4
            -\staccato
            g'4
            -\staccato
            a'4
            -\staccato
        }

        .. figure:: ../_images/image-LeafDynMaker-7.png

    ..  container:: example

        Similarly to Abjad's native classes, it accepts many types of elements
        in its input lists:

        >>> pitches = [0,
        ...            "d'",
        ...            'E4',
        ...            abjad.NumberedPitch(5),
        ...            abjad.NamedPitch("g'"),
        ...            abjad.NamedPitch("A4"),
        ...            ]
        >>> durations = [(1, 32),
        ...              "2/32",
        ...              abjad.Duration("3/32"),
        ...              abjad.Duration(0.125),
        ...              abjad.Duration(5, 32),
        ...              abjad.Duration(6/32),
        ...              ]
        >>> dynamics = ['p',
        ...             abjad.Dynamic('f'),
        ...             ]
        >>> articulations = ['>',
        ...                 abjad.Articulation('-'),
        ...                 abjad.Staccato(),
        ...                 ]
        >>> leaf_dyn_maker = auxjad.LeafDynMaker()
        >>> notes = leaf_dyn_maker(pitches, durations, dynamics, articulations)
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'32
            \p
            - \accent
            d'16
            \f
            - \tenuto
            e'16.
            \staccato
            f'8
            g'8
            ~
            g'32
            a'8.
        }

        .. figure:: ../_images/image-LeafDynMaker-8.png
    """

    ### SPECIAL METHODS ###

    def __call__(self,
                 pitches,
                 durations,
                 dynamics=None,
                 articulations=None,
                 *,
                 omit_repeated_dynamics=False,
                 cyclic_dynamics=False,
                 cyclic_articulations=False,
                 ) -> abjad.Selection:

        r"""Calls the leaf-maker on ``pitches``, ``durations``, ``dynamics``,
        and ``articulations``, returning an ``abjad.Selection``.
        """
        if dynamics:
            for dynamic in dynamics:
                if dynamic is not None:
                    if not isinstance(dynamic, (str, abjad.Dynamic)):
                        raise TypeError("dynamics must be 'str' or "
                                        "'abjad.Dynamic'")
        if articulations:
            for articulation in articulations:
                if articulation is not None:
                    if not isinstance(articulation, (str,
                                                     abjad.Articulation,
                                                     abjad.Staccatissimo,
                                                     abjad.Staccato,
                                                     )):
                        raise TypeError("articulations must be 'str' or "
                                        "'abjad.Articulation'")
        if not isinstance(omit_repeated_dynamics, bool):
            raise TypeError("'omit_repeated_dynamics' must be 'bool")
        if not isinstance(cyclic_dynamics, bool):
            raise TypeError("'cyclic_dynamics' must be 'bool")
        if not isinstance(cyclic_articulations, bool):
            raise TypeError("'cyclic_articulations' must be 'bool")

        leaves = super().__call__(pitches, durations)
        dummy_container = abjad.Container(leaves)
        logical_ties = leaves.logical_ties()

        pitches_ = self._listify(pitches)
        durations_ =  self._listify(durations)
        dynamics_ = self._listify(dynamics)
        articulations_ = self._listify(articulations)

        greatest_len = max(len(pitches_), len(durations_))
        self._fill_list(dynamics_,
                        greatest_len,
                        cyclic=cyclic_dynamics,
                        )
        self._fill_list(articulations_,
                        greatest_len,
                        cyclic=cyclic_articulations,
                        )

        previous_dynamic = None
        for logical_tie, dynamic, articulation in zip(logical_ties,
                                                      dynamics_,
                                                      articulations_):
            if (dynamic and (not omit_repeated_dynamics
                    or dynamic != previous_dynamic)):
                abjad.attach(abjad.Dynamic(dynamic), logical_tie.head)
                previous_dynamic = dynamic
            if articulation and isinstance(articulation, str):
                abjad.attach(abjad.Articulation(articulation),
                             logical_tie.head,
                             )
            elif articulation:
                abjad.attach(articulation, logical_tie.head)

        result = dummy_container[:]
        dummy_container[:] = []

        return result

    ### PRIVATE METHODS ###

    @staticmethod
    def _listify(argument):
        r'Returns a list if argument is not a list.'
        if argument:
            if isinstance(argument, list):
                return argument
            else:
                return [argument]
        else:
            return []

    @staticmethod
    def _fill_list(input_list,
                   length: int,
                   cyclic: bool = False,
                   default=None):
        r"""Extends a list to a certain length, filling it with a default
        value. If ``cyclic`` is ``True``, then it fills the list by cycling the
        original ``input_list``.
        """
        if not cyclic:
            while len(input_list) < length:
                input_list.append(default)
        else:
            original_length = len(input_list)
            while len(input_list) < length:
                current_length = len(input_list)
                input_list.append(input_list[current_length % original_length])



import copy
import random
import abjad


class _AdriftContainer():
    r"""xxx
    """

    ### INITIALISER ###

    def __init__(self,
                 container: abjad.Container,
                 *,
                 head_position: int = 0,
                 ):
        self._initial_container = copy.deepcopy(container)

    ### SPECIAL METHODS ###

    def __call__(self):
        r"""Calls the drifting process for one iteration, returning an
        ``abjad.Selection``.
        """
        self._fade_process()
        return self.current_container

    ### PUBLIC METHODS ###

    def output_all(self):
        r'xxx'
        pass

    ### PRIVATE METHODS ###

    def _fade_process(self):
        pass

    def _remove_random_leaf(self):
        pass

    def _add_random_leaf(self):
        pass

    ### PUBLIC PROPERTIES ###

    @property
    def current_container(self):
        r"""Read-only property, returns the index of the previously output
        element.
        """
        return copy.deepcopy(self._current_container)



import abjad
import pytest
import auxjad


def test_respell_container_01():
    container = abjad.Container(r"c'4 r4 <ef' e'>4 g'4 <c' cs'>4 r2.")
    auxjad.respell_container(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            c'4
            r4
            <ds' e'>4
            g'4
            <c' df'>4
            r2.
        }
        """)


def test_respell_container_02():
    staff = abjad.Staff()
    for pitch in range(12):
        staff.append(abjad.Chord([pitch, pitch + 1], (1, 16)))
    auxjad.respell_container(staff)
    assert format(staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            <c' df'>16
            <cs' d'>16
            <d' ef'>16
            <ds' e'>16
            <e' f'>16
            <f' gf'>16
            <fs' g'>16
            <g' af'>16
            <gs' a'>16
            <a' bf'>16
            <as' b'>16
            <b' c''>16
        }
        """)


def test_respell_container_03():
    container = abjad.Container(r"<a c' cs' f'>1")
    auxjad.respell_container(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            <a c' df' f'>1
        }
        """)


def test_respell_container_04():
    container = abjad.Container(r"<e' cs' g' ef'>1")
    auxjad.respell_container(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            <cs' ds' e' g'>1
        }
        """)


def test_respell_container_05():
    container = abjad.Container(r"<c' cs''>1")
    auxjad.respell_container(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            <c' cs''>1
        }
        """)
    auxjad.respell_container(container, include_multiples=True)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            <c' df''>1
        }
        """)


def test_respell_container_06():
    container = abjad.Container(r"<c' cs' cs''>1")
    auxjad.respell_container(container)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            <c' df' cs''>1
        }
        """)
    container = abjad.Container(r"<c' cs' cs''>1")
    auxjad.respell_container(container, respell_by_pitch_class=True)
    assert format(container) == abjad.String.normalize(
        r"""
        {
            <c' df' df''>1
        }
        """)



import random


class CartographySelector():
    r"""A selector used to store, manipulate, and select objects using a
    weighted function constructed with a fixed decay rate. The decay rate
    represents the ratio of probabilities of any index given the probability of
    the preceeding one. For instance, if the decay rate is set to ``0.75``
    (which is its default value), the probability of the element in index 1 of
    the input list being selected is 0.75 the probability of the element in
    index 0, and the probability of the element in index 2 is 0.5625 (0.75^2)
    the probability of the element in index 0. This is the selector used in my
    *Cartography* series of compositions.

    ..  container:: example

        The selector should be initialised with a list of objects. The contents
        of the list can be absolutely anything.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]

        The default decay rate is 0.75; that is, the weight of any given
        elements is the weight of the previous one multiplied by 0.75. The
        weights are associated with the index position, not the elements
        themselves.

        >>> selector.weights
        [1.0, 0.75, 0.5625, 0.421875, 0.31640625]

        Applying the ``len()`` function to the selector will give the length
        of the input list.

        >>> len(selector)
        5

        Calling the selector will output one of its elements, selected
        according to its weight function.

        >>> selector()
        2

        Alternatively, use the ``next()`` function or ``__next__()`` method to
        get the next result.

        >>> selector.__next__()
        1
        >>> next(selector)
        0

    ..  container:: example

        By default, only the weight function (defined by the decay rate) is
        taken into consideration when selecting an element. This means that
        repeated elements can appear, as shown below.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(selector())
        >>> result
        203001402200011111101400310140

        Calling the selector with the optional keyword argument ``no_repeat``
        set to ``True`` will forbid immediate repetitions among consecutive
        calls.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(selector(no_repeat=True))
        >>> result
        210421021020304024230120241202

    ..  container:: example

        The keyword argument ``decay_rate`` can be used to set a different
        decay rate when creating a selector.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4],
        ...                                       decay_rate=0.5,
        ...                                       )
        >>> selector.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625]

        The decay rate can also be set after the creation of a selector using,
        the property ``decay_rate``.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.decay_rate = 0.2
        >>> selector.weights
        [1.0, 0.2, 0.04000000000000001, 0.008000000000000002,
        0.0016000000000000003]
        >>> result = ''
        >>> for _ in range(30):
        ...     result += str(selector())
        >>> result
        '000001002100000201001030000100'

    ..  container:: example

        Appending is a type of content transformation. It discards the first
        element of the selector's ``contents``, shifts all others leftwards,
        and then appends the new element to the last index.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.append(5)
        >>> selector.contents
        [1, 2, 3, 4, 5]
        >>> selector.append(42)
        >>> selector.contents
        [2, 3, 4, 5, 42]

    ..  container:: example

        The method ``append_keeping_n()`` is similar to ``append()``, but it
        keeps the first ``n`` elements of ``contents`` untouched. It thus
        discards the n+1-th element, shifts all the next elements one position
        lefwards, and finally appends the new element at the last index.

        >>> selector = auxjad.CartographySelector([10, 7, 14, 31, 98])
        >>> selector.contents
        [10, 7, 14, 31, 98]
        >>> selector.append_keeping_n(100, 2)
        >>> selector.contents
        [10, 7, 31, 98, 100]

    ..  container:: example

        Prepending is another type of content transformation. It discards the
        last element of the ``contents``, shifts all others rightwards, and
        then prepends the new element to the first index.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.prepend(-1)
        >>> selector.contents
        [-1, 0, 1, 2, 3]
        >>> selector.prepend(71)
        >>> selector.contents
        [71, -1, 0, 1, 2]

    ..  container:: example

        Rotation is another type of content transformation. It rotates all
        elements rightwards, moving the last element to the first index. If the
        optional keyword argument ``anticlockwise`` is set to ``True``, the
        rotation will be in the opposite direction.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.rotate()
        >>> selector.contents
        [1, 2, 3, 4, 0]
        >>> selector.rotate(anticlockwise=True)
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.rotate(anticlockwise=True)
        >>> selector.contents
        [1, 2, 3, 4, 0]

    ..  container:: example

        The mirror transformation swaps the element of the input index with its
        complementary element. Complementary elements are defined as the pair
        of elements which share the same distance from the centre of the
        ``contents`` (in terms of number of indeces), and are located at either
        side.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.mirror(0)
        >>> selector.contents
        [4, 1, 2, 3, 0]
        >>> selector.mirror(0)
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.mirror(3)
        >>> selector.contents
        [0, 3, 2, 1, 4]
        >>> selector.mirror(2)
        >>> selector.contents
        [0, 3, 2, 1, 4]

    ..  container:: example

        To mirror a random pair of complementary elements, use the
        ``mirror_random()`` method. In case of a selector with an odd number
        of elements, this method will never pick an element at the pivot point
        since the operation would not change the contents.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.mirror_random()
        >>> selector.contents
        [4, 1, 2, 3, 0]
        >>> selector.mirror_random()
        >>> selector.contents
        [4, 3, 2, 1, 0]
        >>> selector.mirror_random()
        >>> selector.contents
        [4, 1, 2, 3, 0]

    ..  container:: example

        The method ``randomise()`` will randomise the position of the elements
        of a selector's ``contents``.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4])
        >>> selector.contents
        [0, 1, 2, 3, 4]
        >>> selector.randomise()
        >>> selector.contents
        [1, 4, 3, 0, 2]

    ..  container:: example

        The contents of a selector can also be altered after it has been
        initialised using the ``contents`` property. The length of the
        contents can change as well.

        >>> selector = auxjad.CartographySelector([0, 1, 2, 3, 4],
        ...                                       decay_rate=0.5,
        ...                                       )
        >>> len(selector)
        5
        >>> selector.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625]
        >>> selector.contents = [10, 7, 14, 31, 98, 47, 32]
        >>> selector.contents
        [10, 7, 14, 31, 98, 47, 32]
        >>> len(selector)
        7
        >>> selector.weights
        [1.0, 0.5, 0.25, 0.125, 0.0625, 0.03125, 0.015625]

    ..  container:: example

        Use the read-only properties ``previous_result`` and ``previous_index``
        to output the previous result and its index.

        >>> selector = auxjad.CartographySelector([10, 7, 14, 31, 98])
        >>> selector()
        14
        >>> previous_index = selector.previous_index
        >>> previous_index
        2
        >>> selector.previous_result
        14

    ..  container:: example

        This class allows indecing and slicing just like regular lists. This
        can be used to both access and alter elements.

        >>> selector = auxjad.CartographySelector([10, 7, 14, 31, 98])
        >>> selector[1]
        7
        >>> selector[1:4]
        [7, 14, 31]
        >>> selector[:]
        [10, 7, 14, 31, 98]
        >>> selector()
        31
        >>> previous_index = selector.previous_index
        >>> previous_index
        3
        >>> selector[previous_index]
        31
        >>> selector.contents
        [10, 7, 14, 31, 98]
        >>> selector[2] = 100
        >>> selector.contents
        [10, 7, 100, 31, 98]
    """

    ### INITIALISER ###

    def __init__(self,
                 contents: list,
                 *,
                 decay_rate: float = 0.75,
                 ):
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list'")
        if not isinstance(decay_rate, float):
            raise TypeError("'decay_rate' must be 'float'")
        if decay_rate <= 0.0 or decay_rate > 1.0:
            raise ValueError("'decay_rate' must be larger than 0.0 and "
                             "less than or equal to 1.0")
        self._contents = contents[:]
        self._previous_index = None
        self._decay_rate = decay_rate
        self._generate_weights()

    ### SPECIAL METHODS ###

    def __repr__(self) -> str:
        r'Outputs the representation of ``contents``.'
        return str(self._contents)

    def __call__(self,
                 *,
                 no_repeat: bool = False,
                 ):
        r'Call the selection process and outputs one element of ``contents``.'
        if not isinstance(no_repeat, bool):
            raise TypeError("'no_repeat' must be 'bool")
        if not no_repeat:
            new_index = random.choices(
                [n for n in range(self.__len__())],
                weights=self.weights,
            )[0]
        else:
            new_index = self._previous_index
            while (new_index == self._previous_index):
                new_index = random.choices(
                    [n for n in range(self.__len__())],
                    weights=self.weights,
                )[0]
        self._previous_index = new_index
        return self._contents[self._previous_index]

    def __next__(self,
                 *,
                 no_repeat: bool = False,
                 ):
        r'Call the selection process and outputs one element of ``contents``.'
        return self.__call__(no_repeat=no_repeat)

    def __len__(self) -> int:
        r'Outputs the length of ``contents``.'
        return len(self._contents)

    def __getitem__(self, key: int):
        r"""Implements reading elements of ``contents`` through indexing or
        slicing of instance.
        """
        return self._contents[key]

    def __setitem__(self, key, value):
        r"""Implements writing elements into ``contents`` through indexing or
        slicing of instance.
        """
        self._contents[key] = value

    ### PUBLIC METHODS ###

    def append(self, new_element):
        r"""A type of content transformation, it discards the first element of
        the ``contents``, shifts all others leftwards, and then appends the new
        element to the last index.
        """
        self._contents = self._contents[1:] + [new_element]

    def append_keeping_n(self, new_element, n: int):
        r"""A type of content transformation similar to ``append()``, it keeps
        the first ``n`` elements of ``contents`` untouched, it then discards
        the n+1-th element, shifts all the next elements one position lefwards,
        and finally appends the new element at the last index.
        """
        self._contents = (self._contents[:n]
                          + self._contents[n+1:]
                          + [new_element])

    def prepend(self, new_element):
        r"""A type of content transformation, it discards the last element of
        the ``contents``, shifts all others rightwards, and then prepends the
        new element to the first index.
        """
        self._contents = [new_element] + self._contents[:-1]

    def rotate(self, anticlockwise=False):
        r"""A type of content transformation, it rotates all elements
        rightwards, moving the last element to the first index. If the optional
        keyword argument ``anticlockwise`` is set to ``True``, the rotation
        will be in the opposite direction.
        """
        if not anticlockwise:
            self._contents = self._contents[1:] + self._contents[:1]
        else:
            self._contents = self._contents[-1:] + self._contents[:-1]

    def mirror(self, index: int):
        r"""A type of content transformation which swaps the element of the
        input index with its complementary element. Complementary elements are
        defined as the pair of elements which share the same distance from the
        centre of the ``contents`` (in terms of number of indeces), and are
        located at either side.
        """
        aux = self._contents[index]
        self._contents[index] = self._contents[-1 - index]
        self._contents[-1 - index] = aux

    def mirror_random(self):
        r"""A type of content transformation which swaps the element of a
        random index with its complementary element. Complementary elements are
        defined as the pair of elements which share the same distance from the
        centre of the ``contents`` (in terms of number of indeces), and are
        located at either side.
        """
        max_index = self.__len__() // 2 - 1
        self.mirror(random.randint(0, max_index))

    def randomise(self):
        r'Randomises the position of the elements of ``contents``.'
        random.shuffle(self._contents)

    ### PRIVATE METHODS ###

    def _generate_weights(self):
        r"""Given a decay rate, this method generates the weights of individual
        indeces.
        """
        self.weights = []
        for n in range(self.__len__()):
            self.weights.append(self._decay_rate ** n)

    ### PUBLIC PROPERTIES ###

    @property
    def contents(self) -> list:
        r'The ``list`` from which the selector picks elements.'
        return self._contents

    @contents.setter
    def contents(self,
                 contents: list
                 ):
        if not isinstance(contents, list):
            raise TypeError("'contents' must be 'list")
        self._contents = contents[:]
        self._generate_weights()

    @property
    def decay_rate(self) -> float:
        r"""The decay rate represents the ratio of probabilities of any index
        given the probability of the preceeding one. For instance, if the decay
        rate is set to ``0.75`` (which is its default value), the probability
        of the element in index 1 of the input list being selected is 0.75 the
        probability of the element in index 0, and the probability of the
        element in index 2 is 0.5625 (0.75^2) the probability of the element in
        index 0.
        """
        return self._decay_rate

    @decay_rate.setter
    def decay_rate(self,
                   decay_rate: float):
        if not isinstance(decay_rate, float):
            raise TypeError("'decay_rate' must be float")
        if decay_rate <= 0.0 or decay_rate > 1.0:
            raise ValueError("'decay_rate' must be larger than 0.0 and less "
                             "than or equal to 1.0")
        self._decay_rate = decay_rate
        self._generate_weights()

    @property
    def previous_index(self):
        r"""Read-only property, returns the index of the previously output
        element.
        """
        return self._previous_index

    @property
    def previous_result(self):
        r'Read-only property, returns the previously output element.'
        return self._contents[self._previous_index]



import abjad


class _HarmonicParent:
    r"""This is the parent class of the ArtificialHarmonic and HarmonicNote
    classes. It implements two common properties with @property and @setter
    decorators.
    """

    ### PUBLIC PROPERTIES ###

    @property
    def direction(self) -> (str, abjad.enums.VerticalAlignment):
        r'The direction of the harmonic note head.'
        return self._direction

    @direction.setter
    def direction(self,
                  direction: (str, abjad.enums.VerticalAlignment),
                  ):
        if direction:
            if not isinstance(direction, (str, abjad.enums.VerticalAlignment)):
                raise TypeError("'direction' must be 'str', None, or either "
                                "'abjad.Up' or 'abjad.Down'")
        self._direction = direction
        markup = self._markup
        self.markup = None
        self.markup = markup

    @property
    def markup(self) -> str:
        r'The markup of the harmonic note head.'
        return self._markup

    @markup.setter
    def markup(self,
               markup: str,
               ):
        if markup:
            if not isinstance(markup, str):
                raise TypeError("'markup' must be 'str'")
            self._markup = markup
            markup = abjad.Markup(self._markup,
                                  direction=self._direction,
                                  )
            abjad.attach(markup, self)
        else:
            self._markup = markup
            markup = abjad.inspect(self).indicator(abjad.Markup)
            if markup:
                abjad.detach(abjad.Markup, self)



import abjad
from ._HarmonicParent import _HarmonicParent


class HarmonicNote(abjad.Note, _HarmonicParent):
    r"""Creates a note with tweaked notehead for harmonics. This is a child
    class of ``abjad.Note``.

    ..  container:: example

        Usage is similar to ``abjad.Note``:

        >>> harm = auxjad.HarmonicNote("c''4")
        >>> harm.style
        'harmonic'
        >>> abjad.f(harm)
        \tweak style #'harmonic
        c''4

        .. figure:: ../_images/image-HarmonicNote-1.png

        And similarly to ``abjad.Note``, pitch and duration can be input in
        many different ways:

        >>> harm1 = auxjad.HarmonicNote("c''4")
        >>> harm2 = auxjad.HarmonicNote("c''", 1/4)
        >>> harm3 = auxjad.HarmonicNote(12, 0.25)
        >>> harm4 = auxjad.HarmonicNote(12, abjad.Duration(1, 4))
        >>> staff = abjad.Staff([harm1, harm2, harm3, harm4])
        >>> abjad.f(staff)
        \new Staff
        {
            \tweak style #'harmonic
            c''4
            \tweak style #'harmonic
            c''4
            \tweak style #'harmonic
            c''4
            \tweak style #'harmonic
            c''4
        }

        .. figure:: ../_images/image-HarmonicNote-2.png

    ..  container:: example

        When creating an ``HarmonicNote``, use the keyword argument ``style``
        to set a different type of note head, such as ``'harmonic-mixed'``:

        >>> harm = auxjad.HarmonicNote("c''4",
        ...                            style='harmonic-mixed',
        ...                            )
        >>> harm.style
        'harmonic-mixed'
        >>> abjad.f(harm)
        \tweak style #'harmonic-mixed
        c''4

        .. figure:: ../_images/image-HarmonicNote-3.png

    ..  container:: example

        Similarly to ``abjad.Note``, ``HarmonicNote`` can take multipliers:

        >>> harm = auxjad.HarmonicNote("c''4",
        ...                            multiplier=(2, 3),
        ...                            )
        >>> harm.multiplier
        abjad.Multiplier(2, 3)
        >>> abjad.f(harm)
        \tweak style #'harmonic
        c''4 * 2/3

        .. figure:: ../_images/image-HarmonicNote-4.png

    ..  container:: example

        All properties of ``abjad.Note`` are also available to be read. This
        class also includes a new property named ``style``:

        >>> harm = auxjad.HarmonicNote("c''4")
        >>> harm.written_pitch
        "c''"
        >>> harm.written_duration
        1/4
        >>> harm.style
        'harmonic'

        All these properties can be set to different values after
        initialisation:

        >>> harm.written_pitch = 18
        >>> harm.written_duration = abjad.Duration(1, 8)
        >>> harm.style = 'harmonic-mixed'
        >>> harm.written_pitch
        "fs''"
        >>> harm.written_duration
        1/8
        >>> harm.style
        'harmonic-mixed'

    ..  container:: example

        To create a harmonic note with a regular note head and with a flageolet
        circle above it, use the style ``'flageolet'``:

        >>> harm = auxjad.HarmonicNote("c''1",
        ...                            style='flageolet',
        ...                            )
        >>> harm.style
        'flageolet'
        >>> abjad.f(harm)
        c''1
        \flageolet

        .. figure:: ../_images/image-HarmonicNote-5.png

    ..  container:: example

        To add a markup expression to the harmonic note, use the markup:

        >>> harm1 = auxjad.HarmonicNote("d''1")
        >>> harm2 = auxjad.HarmonicNote("d''1",
        ...                             markup='III.',
        ...                             )
        >>> harm3 = auxjad.HarmonicNote("d''1",
        ...                             markup='III.',
        ...                             direction=abjad.Down)
        >>> staff = abjad.Staff([harm1, harm2, harm3])
        >>> abjad.f(staff)
        \new Staff
        {
            \tweak style #'harmonic
            d''1
            \tweak style #'harmonic
            d''1
            ^ \markup { III. }
            \tweak style #'harmonic
            d''1
            _ \markup { III. }
        }

        .. figure:: ../_images/image-HarmonicNote-6.png

        Setting ``markup`` to ``None`` will remove the markup from the note.

        >>> harm = auxjad.HarmonicNote("d''1",
        ...                            markup='III.',
        ...                            )
        >>> harm.markup = None
        >>> abjad.f(harm)
        \tweak style #'harmonic
        d''1

        .. figure:: ../_images/image-HarmonicNote-7.png

    ..  warning::

        If another markup is attached to the harmonic note, trying to set the
        ``markup`` to ``None`` will raise an Exception:

        >>> harm = auxjad.HarmonicNote("d''1")
        >>> abjad.attach(abjad.Markup('test'), harm)
        >>> harm.markup = 'III.'
        >>> harm.markup = None
        Exception: multiple indicators attached to client.
    """

    ### INITIALISER ###

    def __init__(self,
                 *arguments,
                 multiplier: abjad.typings.DurationTyping = None,
                 tag: abjad.Tag = None,
                 style: str = 'harmonic',
                 markup: str = None,
                 direction: (str, abjad.enums.VerticalAlignment) = 'up',
                 ):
        super().__init__(*arguments, multiplier=multiplier, tag=tag)
        self.style = style
        self._direction = direction
        self.markup = markup

    ### PUBLIC PROPERTIES ###

    @property
    def style(self) -> str:
        r'The style of the harmonic note head.'
        return self._style

    @style.setter
    def style(self,
              style: str,
              ):
        if not isinstance(style, str):
            raise TypeError("'style' must be 'str'")
        self._style = style
        if not self._style == 'flageolet':
            abjad.tweak(self.note_head).style = self._style
        else:
            flageolet = abjad.LilyPondLiteral(r'\flageolet',
                                              format_slot='after',
                                              )
            abjad.attach(flageolet, self)



import abjad
import pytest
import auxjad


def test_repeat_container_01():
    container = abjad.Container(r"c'4 d'4 e'4")
    output_container = auxjad.repeat_container(container, 3)
    assert format(output_container) == abjad.String.normalize(
        r"""
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }
        """)
    staff = abjad.Staff([output_container])
    assert format(output_container) == abjad.String.normalize(
        r"""
        {
            \time 3/4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }
        """)


def test_repeat_container_02():
    container = abjad.Container(r"\time 3/4 c'2. \time 2/4 r2 g'2")
    output_container = auxjad.repeat_container(container, 3)
    assert format(output_container) == abjad.String.normalize(
        r"""
        {
            %%% \time 3/4 %%%
            c'2.
            %%% \time 2/4 %%%
            r2
            g'2
            %%% \time 3/4 %%%
            c'2.
            %%% \time 2/4 %%%
            r2
            g'2
            %%% \time 3/4 %%%
            c'2.
            %%% \time 2/4 %%%
            r2
            g'2
        }
        """)


def test_repeat_container_03():
    container = abjad.Container(r"\time 3/4 c'4 d'4 e'4 f'2")
    output_container = auxjad.repeat_container(container, 2)
    assert format(output_container) == abjad.String.normalize(
        r"""
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            %%% \time 2/4 %%%
            f'2
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            %%% \time 2/4 %%%
            f'2
        }
        """)


def test_repeat_container_04():
    container = abjad.Container(r"c'4 d'4 e'4")
    output_container = auxjad.repeat_container(container,
                                               3,
                                               omit_all_time_signatures=True,
                                               )
    assert format(output_container) == abjad.String.normalize(
        r"""
        {
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }
        """)


def test_repeat_container_05():
    container = abjad.Container(r"\time 5/4 c'2. d'4 e'4")
    output_container = auxjad.repeat_container(
        container,
        3,
        force_identical_time_signatures=True,
    )
    assert format(output_container) == abjad.String.normalize(
        r"""
        {
            %%% \time 5/4 %%%
            c'2.
            d'4
            e'4
            %%% \time 5/4 %%%
            c'2.
            d'4
            e'4
            %%% \time 5/4 %%%
            c'2.
            d'4
            e'4
        }
        """)


def test_repeat_container_06():
    container = abjad.Staff(r"c'4 d'4 e'4")
    output_staff = auxjad.repeat_container(container, 3)
    assert format(output_staff) == abjad.String.normalize(
        r"""
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
            c'4
            d'4
            e'4
        }
        """)


def test_repeat_container_07():
    container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
    with pytest.raises(ValueError):
        assert auxjad.repeat_container(container, 4)



import copy
import random
import abjad
from ..utilities.leaves_are_tieable import leaves_are_tieable


class _LoopParent():
    r"""This is the parent class of all LoopByXxxx classes. It implements all
    common methods and properties, and initialises those using their @setter
    methods.
    """

    ### INITIALISER ###

    def __init__(self,
                 head_position,
                 window_size,
                 step_size,
                 max_steps,
                 repetition_chance,
                 forward_bias,
                 move_window_on_first_call,
                 ):
        if not isinstance(move_window_on_first_call, bool):
            raise TypeError("'move_window_on_first_call' must be 'bool'")
        self.head_position = head_position
        self.window_size = window_size
        self.step_size = step_size
        self.max_steps = max_steps
        self.repetition_chance = repetition_chance
        self.forward_bias = forward_bias
        self._is_first_window = not move_window_on_first_call

    ### SPECIAL METHODS ###

    def __call__(self) -> abjad.Selection:
        r"""Calls the looping process for one iteration, returning an
        ``abjad.Selection``.
        """
        self._move_head()
        if self._done:
            raise RuntimeError("'contents' has been exhausted")
        self._slice_contents()
        return self.current_window

    def __iter__(self):
        r'Allows instances to be used as iterators.'
        return self

    def __next__(self) -> abjad.Selection:
        r"""Calls the looping process for one iteration, returning an
        ``abjad.Selection``.
        """
        self._move_head()
        if self._done:
            raise StopIteration
        self._slice_contents()
        return self._current_window

    ### PUBLIC METHODS ###

    def output_all(self,
                   *,
                   tie_identical_pitches: bool = False,
                   ) -> abjad.Selection:
        r"""Goes through the whole looping process and outputs a single
        ``abjad.Selection``.
        """
        if not isinstance(tie_identical_pitches, bool):
            raise TypeError("'tie_identical_pitches' must be 'bool'")
        dummy_container = abjad.Container()
        while True:
            try:
                if not tie_identical_pitches or len(dummy_container) == 0:
                    dummy_container.append(self.__call__())
                else:
                    new_window = self.__call__()
                    leaf1 = abjad.select(new_window).leaves()[0]
                    leaf2 = abjad.select(dummy_container).leaves()[-1]
                    if leaves_are_tieable(leaf1, leaf2):
                        abjad.attach(abjad.Tie(), dummy_container[-1])
                    dummy_container.append(new_window)
            except RuntimeError:
                break
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    def output_n(self,
                 n: int,
                 *,
                 tie_identical_pitches: bool = False,
                 ) -> abjad.Selection:
        r"""Goes through ``n`` iterations of the looping process and outputs a
        single ``abjad.Selection``.
        """
        if not isinstance(n, int):
            raise TypeError("first positional argument must be 'int'")
        if n < 1:
            raise ValueError("first positional argument must be a positive "
                             "'int'")
        if not isinstance(tie_identical_pitches, bool):
            raise TypeError("'tie_identical_pitches' must be 'bool'")
        dummy_container = abjad.Container()
        for _ in range(n):
            if not tie_identical_pitches or len(dummy_container) == 0:
                dummy_container.append(self.__call__())
            else:
                new_window = self.__call__()
                leaf1 = abjad.select(new_window).leaves()[0]
                leaf2 = abjad.select(dummy_container).leaves()[-1]
                if leaves_are_tieable(leaf1, leaf2):
                    abjad.attach(abjad.Tie(), dummy_container[-1])
                dummy_container.append(new_window)
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    ### PRIVATE METHODS ###

    def _move_head(self):
        r"""Moves the head by a certain number of steps of fixed size, either
        forwards or backwards according to the forward bias
        """
        if not self._is_first_window:  # 1st window always at initial position
            if (self._repetition_chance == 0.0
                    or random.random() > self._repetition_chance):
                step = self._step_size * random.randint(1, self._max_steps)
                diretion = self._biased_choice(self._forward_bias)
                self._head_position += step * diretion
        else:
            self._is_first_window = False

    def _slice_contents(self):
        pass

    @staticmethod
    def _biased_choice(bias):
        return random.choices([1, -1], weights=[bias, 1.0-bias])[0]

    @staticmethod
    def _remove_all_time_signatures(contents):
        for leaf in abjad.select(contents).leaves():
            if abjad.inspect(leaf).effective(abjad.TimeSignature):
                abjad.detach(abjad.TimeSignature, leaf)

    ### PUBLIC PROPERTIES ###

    @property
    def head_position(self) -> int:
        r'The position of the head at the start of a looping window.'
        return self._head_position

    @head_position.setter
    def head_position(self,
                      head_position: int,
                      ):
        if not isinstance(head_position, int):
            raise TypeError("'head_position' must be 'int'")
        if head_position < 0:
            raise ValueError("'head_position' must be a positive 'int'")
        if head_position >= self._contents.__len__():
            raise ValueError("'head_position' must be smaller than the length "
                             "of 'contents'")
        self._is_first_window = True
        self._head_position = head_position

    @property
    def window_size(self) -> int:
        r'The length of the looping window.'
        return self._window_size

    @window_size.setter
    def window_size(self,
                    window_size: int,
                    ):
        if not isinstance(window_size, int):
            raise TypeError("'window_size' must be 'int'")
        if window_size < 1:
            raise ValueError("'window_size' must be greater than zero")
        if window_size > self._contents.__len__():
            raise ValueError("'window_size' must be smaller than or equal to "
                             "the length of 'contents'")
        self._window_size = window_size

    @property
    def step_size(self) -> int:
        r'The size of each step when moving the head.'
        return self._step_size

    @step_size.setter
    def step_size(self,
                  step_size: int,
                  ):
        if not isinstance(step_size, int):
            raise TypeError("'step_size' must be 'int'")
        if step_size < 1:
            raise ValueError("'step_size' must be greater than zero")
        if step_size >= self._contents.__len__():
            raise ValueError("'step_size' must be smaller than the length of "
                             "'contents'")
        self._step_size = step_size

    @property
    def max_steps(self) -> int:
        r'The maximum number of steps per operation.'
        return self._max_steps

    @max_steps.setter
    def max_steps(self,
                  max_steps: int,
                  ):
        if not isinstance(max_steps, int):
            raise TypeError("'max_steps' must be 'int'")
        if max_steps < 1:
            raise ValueError("'max_steps' must be greater than zero")
        self._max_steps = max_steps

    @property
    def repetition_chance(self) -> float:
        r'The chance of the head not moving, thus repeating the output.'
        return self._repetition_chance

    @repetition_chance.setter
    def repetition_chance(self,
                          repetition_chance: float,
                          ):
        if not isinstance(repetition_chance, float):
            raise TypeError("'repetition_chance' must be 'float'")
        if repetition_chance < 0.0 or repetition_chance > 1.0:
            raise ValueError("'repetition_chance' must be between 0.0 and 1.0")
        self._repetition_chance = repetition_chance

    @property
    def forward_bias(self) -> float:
        r"""The chance of the window moving forward instead of backwards. It
        should range from 0.0 to 1.0 (default 1.0, which means the window can
        only move forwards. A value of 0.5 gives 50% chance of moving forwards
        while a value of 0.0 will move the window only backwards).
        """
        return self._forward_bias

    @forward_bias.setter
    def forward_bias(self,
                     forward_bias: float,
                     ):
        if not isinstance(forward_bias, float):
            raise TypeError("'forward_bias' must be 'float'")
        if forward_bias < 0.0 or forward_bias > 1.0:
            raise ValueError("'forward_bias' must be between 0.0 and 1.0")
        self._forward_bias = forward_bias

    @property
    def current_window(self) -> (list, abjad.Selection):
        r"""Read-only property, returns the window at the current head
        position.
        """
        return copy.deepcopy(self._current_window)

    ### PRIVATE PROPERTIES ###

    @property
    def _done(self) -> bool:
        r"""Boolean indicating whether the process is done (i.e. whether the
        head position has overtaken the ``contents`` length).
        """
        return (self._head_position >= self._contents.__len__()
            or self._head_position < 0)
